/*
    tabs.c
	This part deals with the crontab files.
    
    Copyright 2003 - 2004 Night Productions, by Michael A. Nickerson (aka Darkshadow).  All rights reserved.
    Licensed under the Open Software License version 2.0
*/
#include <sys/stat.h>
#include <stddef.h>
#include <pwd.h>
#include <ctype.h>
#include "loop.h"

#ifdef HAVE_DIRENT_H
    #include <dirent.h>
#endif
#ifdef HAVE_NDIR_H
    #include <ndir.h>
#endif
#ifdef HAVE_SYS_DIR_H
    #include <sys/dir.h>
#endif
#ifdef HAVE_SYS_NDIR_H
    #include <sys/ndir.h>
#endif

#ifdef HAVE_SYSLOG_H
    #include <syslog.h>
#else
    #define LOG_CRIT 0
    #define LOG_INFO 0
    #define LOG_ERR 0
    #define LOG_PID 0
    #define LOG_DAEMON 0
#endif /* defined HAVE_SYSLOG_H */

int parsetab(char *tab, char *words[], int max)
{
    char *p = tab;
    int i = 0;

    while(1) {
	while(isspace(*p)) p++;
	if(*p == '\0')
	    return i;
	words[i++] = p;
	while(!isspace(*p) && *p != '\0')
	    p++;
	if(*p == '\0')
	    return i;
	if(i >= max)
	    return i;
	*p++ = '\0';
    }
}

char *envget(const char *env, char *environment[])
{
    int i = 0;
    size_t length = strlen(env);

    while((environment[i] != NULL)) {
	if ( !(strncmp(environment[i], env, length)) ) {
	    char *t;
	    t = strchr(environment[i],'=');
	    *t++;
	    return t;
	}
	i++;
    }
    return NULL;
}

void envset(const char *env, const char *var, char *environment[])
{
    int i = 0;
    if (envget(env, environment) != NULL) {
	while ((environment[i] != NULL)) {
	    if (!strncmp(environment[i],env,strlen(env))) {
		free(environment[i]);
		environment[i] = (char *)malloc((strlen(env) + strlen(var)) + 3);
		if (environment[i] == NULL) {
		    timeLog(LOG_CRIT, "No system memory available! Exiting...");
		    exit(1);
		}
		sprintf(environment[i],"%s=%s",env,var);
		break;
	    }
	    i++;
	}
    } else {
	while((environment[i] != NULL && i < MAXENVVARS)) {
	    i++;
	}
	environment[i] = (char *)malloc((strlen(env) + strlen(var)) + 3);
	if (environment[i] == NULL) {
	    timeLog(LOG_CRIT,"No system memory available! Exiting...");
	    exit(1);
	}
	sprintf(environment[i],"%s=%s",env,var);
	environment[(i + 1)] = NULL;
    }
}

void envreplace(char *envString, char *env[])
{
    char *copy, *envPtr, *varPtr, *t;
    
    copy = (char *)malloc(strlen(envString) + 1);
    if (copy == NULL) {
	timeLog(LOG_CRIT, "No system memory available! Exiting...");
	exit(1);
    }
    strcpy(copy, envString);
    
    t = strchr(copy, '=');
    *t = '\0';
    *t++;
    envPtr = copy;
    varPtr = t;
    
    envset(envPtr, varPtr, env);
    free(copy);
    t = NULL;
}

void envcpy(char *env1[], char *env2[])
{
    int i = 0;
    while ((env2[i] != NULL && i < MAXENVVARS)) {
	env1[i] = malloc(strlen(env2[i]) + 3);
	if (env1[i] == NULL) {
	    timeLog(LOG_CRIT,"No system memory! Exiting...");
	    exit(1);
	}
	strcpy(env1[i],env2[i]);
	i++;
    }
    env1[i] = NULL;
}

void releaseTabs(struct crontab *theTab)
{
    int i = 0;
    if (theTab->nextTab != NULL) {
      releaseTabs(theTab->nextTab);
      theTab->nextTab = NULL;
    }
    if (theTab->cron_user != NULL) {
	free(theTab->cron_user);
	theTab->cron_user = NULL;
    }
    if (theTab->cron_min != NULL) {
	free(theTab->cron_min);
	theTab->cron_min = NULL;
    }
    if (theTab->cron_hour != NULL) {
	free(theTab->cron_hour);
	theTab->cron_hour = NULL;
    }
    if (theTab->cron_mday != NULL) {
	free(theTab->cron_mday);
	theTab->cron_mday = NULL;
    }
    if (theTab->cron_month != NULL) {
	free(theTab->cron_month);
	theTab->cron_month = NULL;
    }
    if (theTab->cron_wday != NULL) {
	free(theTab->cron_wday);
	theTab->cron_wday = NULL;
    }
    if (theTab->cron_command != NULL) {
	free(theTab->cron_command);
	theTab->cron_command = NULL;
    }
    
    while((theTab->environment[i] != NULL)) {
	free(theTab->environment[i]);
	theTab->environment[i] = NULL;
	i++;
    }
    free(theTab);
    theTab = NULL;
}

void setFileEnv(char *inFile, char *tabEnv[])
{
    char *marker = NULL, *envPtr = NULL, *varPtr = NULL, *t = NULL;
    size_t length = 0, count = 0;
    int inQuote = 0;
    
    if((t = strrchr(inFile,'\n')) != NULL) {
	*t='\0';
    }
    
    length = strlen(inFile);
    for (t = inFile; *t != '\0' && count < length; *t++, count++) {
	if (*t == '=') {
	    *t = '\0';
	    *t++;
	    count++;
	    if (count < length) {
		marker = t;
		break;
	    } else {
		marker = NULL;
		break;
	    }
	}
    }
    envPtr = inFile;
    varPtr = marker;
    if (varPtr == NULL) { varPtr = ""; }
    
    length=strlen(envPtr);
    for (t = envPtr; *t != '\0'; *t++) {
	if ((*t == '\"') && !inQuote) {
	    *t = '\0';
	    inQuote = 1;
	} else if ((*t == '\"') && inQuote) {
	    *t = '\0';
	    inQuote = 0;
	} else if (!inQuote && (*t == ' ')) {
	    *t = '\0';
	}
    }
    
    count = 0;
    while ((!*envPtr && count < length)) {
	*envPtr++;
	count++;
    }
    
    length = strlen(varPtr);
    
    for (t = varPtr; *t != '\0'; *t++) {
	if (!inQuote && (*t == ' ')) {
	    *t = '\0';
	} else if (!inQuote && (*t == '\"')) {
	    *t = '\0';
	    inQuote = 1;
	} else if (inQuote && (*t == '\"')) {
	    *t = '\0';
	    inQuote = 0;
	}
    }
    
    count = 0;
    while ((!*varPtr && count < length)) {
	*varPtr++;
	count++;
    }
    
    envset(envPtr,varPtr,tabEnv);
    t = NULL;
}
		
struct crontab *loadTab(struct crontab *tab, const char *tabFile, const char *userName)
{
    char *inFile;
    char *words[8], *t, *env[MAXENVVARS] = { NULL }, *tabEnv[MAXENVVARS];
    char *spare;
    FILE *cronFile=fopen(tabFile, "r");
    struct crontab *currentTab = tab;
    struct stat ft;
    int i = 0, fileFD = 0;
    int comm = (userName == NULL) ? 6 : 5;
    int args = (userName == NULL) ? 7 : 6;
    
    if (cronFile == NULL) {
	return currentTab;
    }
    
    fileFD = fileno(cronFile);
    
    if (fstat(fileFD, &ft) != 0) {
	return currentTab;
    }
    
    inFile = (char *)malloc(ft.st_size);
    spare = (char *)malloc(ft.st_size);
    
    if ((inFile == NULL) || (spare == NULL)) {
	timeLog(LOG_CRIT, "No system memory available! Exiting...");
	return currentTab;
    }
    
    tabEnv[0] = NULL;
    env[0] = NULL;
    
    endpwent();
    
    while(fgets(inFile, ft.st_size, cronFile) != NULL) {
	uid_t ID = 0;
	gid_t GID = 0;
	char *home = NULL, *shell = NULL;
	struct passwd *pw;
	struct crontab *nextSysTab;
	int next = 0;
	if ( strlen( inFile ) == 0 ) continue;
	if (*inFile == '#') continue;
	if ((strchr(inFile, '=')) != NULL) {
	    setFileEnv(inFile, tabEnv);
	    continue;
	}
	if (!strncmp(inFile, "@reboot", 7)) { continue; }
	else if (!(strncmp(inFile, "@yearly", 7)) || !(strncmp(inFile, "@annually", 9))) {
	    parsetab(inFile, words, 2);
	    sprintf(spare, "0 0 1 1 * %s", words[1]);
	    parsetab(spare, words, args);
	}
	else if (!strncmp(inFile, "@monthly", 8)) {
	    parsetab(inFile, words, 2);
	    sprintf(spare, "0 0 1 * * %s", words[1]);
	    parsetab(spare, words, args);
	}
	else if (!strncmp(inFile, "@weekly", 7)) {
	    parsetab(inFile, words, 2);
	    sprintf(spare, "0 0 * * 0 %s", words[1]);
	    parsetab(spare, words, args);
	}
	else if (!(strncmp(inFile, "@daily", 6)) || !(strncmp(inFile, "@midnight", 9))) {
	    parsetab(inFile, words, 2);
	    sprintf(spare, "0 0 * * * %s", words[1]);
	    parsetab(spare, words, args);
	}
	else if (!strncmp(inFile, "@hourly", 7)) {
	    parsetab(inFile, words, 2);
	    sprintf(spare, "0 * * * * %s", words[1]);
	    parsetab(spare, words, args);
	} else {
	    parsetab(inFile, words, args);
	}
	if ((t = strrchr(words[comm], '\n')) != NULL)
	    *t = '\0';
	t = NULL;
	nextSysTab = malloc(sizeof(struct crontab));
	if (nextSysTab == NULL) {
	    timeLog(LOG_CRIT, "No system memory available! Exiting...");
	    exit(1);
	}
	
	if (userName == NULL) {
	    currentTab->cron_user = (char *)malloc(strlen(words[5]) + 1);
	
	    if (currentTab->cron_user == NULL) {
		timeLog(LOG_CRIT, "No system memory available! Exiting...");
		exit(1);
	    }
	
	    strcpy(currentTab->cron_user, words[5]);
	
	} else {
	    currentTab->cron_user = (char *)malloc(strlen(userName) + 1);
	    
	    if (currentTab->cron_user == NULL) {
		timeLog(LOG_CRIT, "No system memory available! Exiting...");
		exit(1);
	    }
	    
	    strcpy(currentTab->cron_user, userName);
	}
	
	pw = getpwnam(currentTab->cron_user);
	
	if (pw != NULL) {
	    ID = pw->pw_uid;
	    GID = pw->pw_gid;
	    home = pw->pw_dir;
	    shell = pw->pw_shell;
	} else {
	    home = "/dev/null";
	    shell = "/bin/sh";
	}
	
	envset("USER", currentTab->cron_user, env);
	envset("LOGNAME", currentTab->cron_user, env);
	envset("HOME", home, env);
	envset("SHELL", shell, env);
	envset("PATH", "/bin:/sbin:/usr/bin:/usr/sbin", env);
	
	while((tabEnv[next] != NULL)) {
	    envreplace(tabEnv[next], env);
	    next++;
	}
	
	envcpy(currentTab->environment, env);
	currentTab->cron_user_uid = ID;
	currentTab->cron_user_gid = GID;
	
	currentTab->cron_min = (char *)malloc(strlen(words[0]) + 1);
	currentTab->cron_hour = (char *)malloc(strlen(words[1]) + 1);
	currentTab->cron_mday = (char *)malloc(strlen(words[2]) + 1);
	currentTab->cron_month = (char *)malloc(strlen(words[3]) + 1);
	currentTab->cron_wday = (char *)malloc(strlen(words[4]) + 1);
	
	if ((currentTab->cron_min == NULL) || (currentTab->cron_hour == NULL) || (currentTab->cron_mday == NULL) || (currentTab->cron_month == NULL) || (currentTab->cron_wday == NULL)) {
	    timeLog(LOG_CRIT, "No system memory available! Exiting...");
	    exit(1);
	}

	strcpy(currentTab->cron_min, words[0]);
	strcpy(currentTab->cron_hour, words[1]);
	strcpy(currentTab->cron_mday, words[2]);
	strcpy(currentTab->cron_month, words[3]);
	strcpy(currentTab->cron_wday, words[4]);
#ifdef HAVE_FRAMEWORKS
	if (!strncmp(words[comm], "@AppleNotOnBattery", 18)) {
	    char *swords[3];
	    parsetab(words[comm], swords, 2);
	    
	    currentTab->cron_command = (char *)malloc(strlen(swords[1]) + 1);
	    
	    if (currentTab->cron_command == NULL) {
		timeLog(LOG_CRIT, "No system memory available! Exiting...");
		exit(1);
	    }
	
	    strcpy(currentTab->cron_command, swords[1]);
		
	    currentTab->batteryFlags = 1;
	} else {
#endif
	    currentTab->cron_command = (char *)malloc(strlen(words[comm]) + 1);
	    
	    if (currentTab->cron_command == NULL) {
		timeLog(LOG_CRIT, "No system memory available! Exiting...");
		exit(1);
	    }
	    
	    strcpy(currentTab->cron_command, words[comm]);
#ifdef HAVE_FRAMEWORKS
	    currentTab->batteryFlags = 0;
	}
#endif
	currentTab->nextTab = nextSysTab;
	currentTab = nextSysTab;
	currentTab->nextTab = NULL;
    }
    fclose(cronFile);
    i = 0;
    while ((env[i] != NULL)) {
	free(env[i]);
	env[i] = NULL;
	i++;
    }
    i=0;
    while ((tabEnv[i] != NULL)) {
	free(tabEnv[i]);
	tabEnv[i] = NULL;
	i++;
    }
    
    if (inFile != NULL) {
	free(inFile);
	inFile = NULL;
    }
    if (spare != NULL) {
	free(spare);
	spare = NULL;
    }
    
    endpwent();
    
    return currentTab;
}

void loadUserTab(struct crontab *tab)
{
    struct crontab *currentTab = tab;
    DIR *crontabdir = NULL;
    struct dirent *dirPtr = NULL;
        
    if ((crontabdir = opendir(userCrontabDirectoryDef)) != NULL) {
	while ((dirPtr = readdir(crontabdir)) != NULL) {
	    char *path;
	    
	    if (dirPtr->d_name[0] == '.') { 
		continue;
	    }
	    
	    path = (char *)malloc((strlen(dirPtr->d_name) + strlen(userCrontabDirectoryDef)) + 1);
	    if (path == NULL) {
		timeLog(LOG_CRIT, "No system memory available! Exiting...");
		exit(1);
	    }
	    
	    sprintf(path,"%s/%s", userCrontabDirectoryDef,dirPtr->d_name);
	    
	    currentTab = loadTab(currentTab, path, dirPtr->d_name);
	    
	    free(path);
	}
    }
    closedir(crontabdir);
}

void loadDatabase(struct cronrep *database)
{
    struct crontab *firstUserTab, *firstSysTab;
    struct stat sysSt, userSt;
    firstUserTab = malloc(sizeof(struct crontab));
    firstSysTab = malloc(sizeof(struct crontab));
    if ((firstUserTab == NULL) || (firstSysTab == NULL)) {
	timeLog(LOG_CRIT,"No system memory available! Exiting...");
	exit(1);
    }
    
    stat(systemCrontabFileDef, &sysSt);
    stat(userCrontabDirectoryDef, &userSt);
    
    firstSysTab->nextTab = NULL;
    firstUserTab->nextTab = NULL;
    
    database->systemTab = firstSysTab;
    database->userTab = firstUserTab;
    database->sysTabModDate = sysSt.st_mtime;
    database->userTabModDate = userSt.st_mtime;
    
    loadUserTab(database->userTab);
    (void)loadTab(database->systemTab, systemCrontabFileDef, NULL);
}

void reloadDatabase(struct cronrep *database, int tab)
{
    struct crontab *aTab = malloc(sizeof(struct crontab));
    if (aTab == NULL) {
	timeLog(LOG_ERR, "No system memory available! Keeping current database...");
	switch (tab) {
	    case SystemTabsDef:
		database->sysTabModDate = 0;
		break;
	    case UserTabsDef:
		database->userTabModDate = 0;
		break;
	    default:
		break;
	}
	return;
    }
    switch (tab) {
	struct stat st;
	case SystemTabsDef:
	    timeLog(LOG_INFO, "Reloading system crontab");
	    stat(systemCrontabFileDef, &st);
	    releaseTabs(database->systemTab);
	    database->systemTab = aTab;
	    database->sysTabModDate = st.st_mtime;
	    (void)loadTab(database->systemTab, systemCrontabFileDef, NULL);
	    break;
	case UserTabsDef:
	    timeLog(LOG_INFO, "Reloading user crontabs");
	    stat(userCrontabDirectoryDef, &st);
	    releaseTabs(database->userTab);
	    database->userTab = aTab;
	    database->userTabModDate = st.st_mtime;
	    loadUserTab(database->userTab);
	    break;
	default:
	    break;
    }
}
